from .Lyric_character import Lyric_character
from .Lyric_file import Lyric_file
from .Lyric_line import Lyric_line
from .Lyric_line_content import Lyric_line_content
from .Lyric_Time_tab import Lyric_Time_tab

name = "MyLyric"

__version__ = "3.0.0"
__description__ = "A simple light Lyric File processing module"
__author__ = "Ding Zedong"
__author_email__ = "2701690963@qq.com"
__all__ = ["Lyric_file", "Lyric_line", "Lyric_Time_tab", "Lyric_line_content", "Lyric_character"]
